@@ -238,10 +238,10 @@ bool eval_time(const struct expr *expr, struct eval_state *state) {
 	switch (expr->time_unit) {
 	case DAYS:
 		diff /= 60*24;
-		fallthrough;
+		BFS_FALLTHROUGH;
 	case MINUTES:
 		diff /= 60;
-		fallthrough;
+		BFS_FALLTHROUGH;
 	case SECONDS:
 		break;
 	}
@@ -644,12 +644,13 @@ bool eval_fls(const struct expr *expr, struct eval_state *state) {
 	}
 
 	uintmax_t ino = statbuf->ino;
-	uintmax_t blocks = ((uintmax_t)statbuf->blocks*BFS_STAT_BLKSIZE + 1023)/1024;
+	uintmax_t block_size = state->ctx->posixly_correct ? 512 : 1024;
+	uintmax_t blocks = ((uintmax_t)statbuf->blocks*BFS_STAT_BLKSIZE + block_size - 1)/block_size;
 	char mode[11];
 	xstrmode(statbuf->mode, mode);
 	char acl = bfs_check_acl(ftwbuf) > 0 ? '+' : ' ';
 	uintmax_t nlink = statbuf->nlink;
-	if (fprintf(file, "%9ju %6ju %s%c %2ju ", ino, blocks, mode, acl, nlink) < 0) {
+	if (fprintf(file, "%9ju %6ju %s%c %2ju", ino, blocks, mode, acl, nlink) < 0) {
 		goto error;
 	}
 
@@ -759,7 +760,7 @@ bool eval_fprint0(const struct expr *expr, struct eval_state *state) {
  * -f?printf action.
  */
 bool eval_fprintf(const struct expr *expr, struct eval_state *state) {
-	if (bfs_printf(expr->cfile->file, expr->printf, state->ftwbuf) != 0) {
+	if (bfs_printf(expr->cfile, expr->printf, state->ftwbuf) != 0) {
 		eval_report_error(state);
 	}
 
@@ -1415,11 +1416,32 @@ done:
 	return state.action;
 }
 
+/** Check if an rlimit value is infinite. */
+static bool rlim_isinf(rlim_t r) {
+	// Consider RLIM_{INFINITY,SAVED_{CUR,MAX}} all equally infinite
+	if (r == RLIM_INFINITY) {
+		return true;
+	}
+
+#ifdef RLIM_SAVED_CUR
+	if (r == RLIM_SAVED_CUR) {
+		return true;
+	}
+#endif
+
+#ifdef RLIM_SAVED_MAX
+	if (r == RLIM_SAVED_MAX) {
+		return true;
+	}
+#endif
+
+	return false;
+}
+
 /** Compare two rlimit values, accounting for RLIM_INFINITY etc. */
 static int rlim_cmp(rlim_t a, rlim_t b) {
-	// Consider RLIM_{INFINITY,SAVED_{CUR,MAX}} all equally infinite
-	bool a_inf = a == RLIM_INFINITY || a == RLIM_SAVED_CUR || a == RLIM_SAVED_MAX;
-	bool b_inf = b == RLIM_INFINITY || b == RLIM_SAVED_CUR || b == RLIM_SAVED_MAX;
+	bool a_inf = rlim_isinf(a);
+	bool b_inf = rlim_isinf(b);
 	if (a_inf || b_inf) {
 		return a_inf - b_inf;
 	}
