@@ -102,7 +102,7 @@ fail:
 	return -1;
 }
 
-struct bfs_mtab *bfs_mtab_parse() {
+struct bfs_mtab *bfs_mtab_parse(void) {
 	struct bfs_mtab *mtab = malloc(sizeof(*mtab));
 	if (!mtab) {
 		return NULL;
@@ -162,7 +162,7 @@ struct bfs_mtab *bfs_mtab_parse() {
 
 #elif BFS_MNTTAB
 
-	FILE *file = fopen(MNTTAB, "r");
+	FILE *file = xfopen(MNTTAB, O_RDONLY | O_CLOEXEC);
 	if (!file) {
 		error = errno;
 		goto fail;
