@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2021 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2022 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -237,10 +237,14 @@ struct parser_state {
 	bool stdout_tty;
 	/** Whether this session is interactive (stdin and stderr are each a terminal). */
 	bool interactive;
+	/** Whether stdin has been consumed by -files0-from -. */
+	bool stdin_consumed;
 	/** Whether -color or -nocolor has been passed. */
 	enum use_color use_color;
 	/** Whether a -print action is implied. */
 	bool implicit_print;
+	/** Whether the default root "." should be used. */
+	bool implicit_root;
 	/** Whether the expression has started. */
 	bool expr_started;
 	/** Whether any non-option arguments have been encountered. */
@@ -260,6 +264,8 @@ struct parser_state {
 	const char *mount_arg;
 	/** An "-xdev"-type argument if any. */
 	const char *xdev_arg;
+	/** An "-ok"-type argument if any. */
+	const char *ok_arg;
 
 	/** The current time. */
 	struct timespec now;
@@ -328,13 +334,39 @@ static void init_print_expr(struct parser_state *state, struct expr *expr) {
 static int expr_open(struct parser_state *state, struct expr *expr, const char *path) {
 	struct bfs_ctx *ctx = state->ctx;
 
-	expr->cfile = bfs_ctx_open(ctx, path, state->use_color);
-	if (!expr->cfile) {
-		parse_error(state, "${blu}%s${rs} ${bld}%s${rs}: %m.\n", expr->argv[0], path);
-		return -1;
+	FILE *file = NULL;
+	CFILE *cfile = NULL;
+
+	file = xfopen(path, O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC);
+	if (!file) {
+		goto fail;
 	}
 
+	cfile = cfwrap(file, state->use_color ? ctx->colors : NULL, true);
+	if (!cfile) {
+		goto fail;
+	}
+
+	CFILE *dedup = bfs_ctx_dedup(ctx, cfile, path);
+	if (!dedup) {
+		goto fail;
+	}
+
+	if (dedup != cfile) {
+		cfclose(cfile);
+	}
+
+	expr->cfile = dedup;
 	return 0;
+
+fail:
+	parse_error(state, "${blu}%s${rs} ${bld}%s${rs}: %m.\n", expr->argv[0], path);
+	if (cfile) {
+		cfclose(cfile);
+	} else if (file) {
+		fclose(file);
+	}
+	return -1;
 }
 
 /**
@@ -383,12 +415,21 @@ static char **parser_advance(struct parser_state *state, enum token_type type, s
  * Parse a root path.
  */
 static int parse_root(struct parser_state *state, const char *path) {
-	struct bfs_ctx *ctx = state->ctx;
-	int ret = DARRAY_PUSH(&ctx->paths, &path);
-	if (ret != 0) {
-		parse_perror(state, "DARRAY_PUSH()");
+	char *copy = strdup(path);
+	if (!copy) {
+		parse_perror(state, "strdup()");
+		return -1;
 	}
-	return ret;
+
+	struct bfs_ctx *ctx = state->ctx;
+	if (DARRAY_PUSH(&ctx->paths, &copy) != 0) {
+		parse_perror(state, "DARRAY_PUSH()");
+		free(copy);
+		return -1;
+	}
+
+	state->implicit_root = false;
+	return 0;
 }
 
 /**
@@ -463,7 +504,11 @@ static const char *parse_int(const struct parser_state *state, const char *str,
 	errno = 0;
 	long long value = strtoll(str, &endptr, base);
 	if (errno != 0) {
-		goto bad;
+		if (errno == ERANGE) {
+			goto range;
+		} else {
+			goto bad;
+		}
 	}
 
 	if (endptr == str) {
@@ -475,20 +520,20 @@ static const char *parse_int(const struct parser_state *state, const char *str,
 	}
 
 	if ((flags & IF_UNSIGNED) && value < 0) {
-		goto bad;
+		goto negative;
 	}
 
 	switch (flags & IF_SIZE_MASK) {
 	case IF_INT:
 		if (value < INT_MIN || value > INT_MAX) {
-			goto bad;
+			goto range;
 		}
 		*(int *)result = value;
 		break;
 
 	case IF_LONG:
 		if (value < LONG_MIN || value > LONG_MAX) {
-			goto bad;
+			goto range;
 		}
 		*(long *)result = value;
 		break;
@@ -509,6 +554,18 @@ bad:
 		parse_error(state, "${bld}%s${rs} is not a valid integer.\n", str);
 	}
 	return NULL;
+
+negative:
+	if (!(flags & IF_QUIET)) {
+		parse_error(state, "Negative integer ${bld}%s${rs} is not allowed here.\n", str);
+	}
+	return NULL;
+
+range:
+	if (!(flags & IF_QUIET)) {
+		parse_error(state, "${bld}%s${rs} is too large an integer.\n", str);
+	}
+	return NULL;
 }
 
 /**
@@ -961,16 +1018,16 @@ static struct expr *parse_time(struct parser_state *state, int field, int arg2)
 		switch (*tail) {
 		case 'w':
 			time *= 7;
-			fallthrough;
+			BFS_FALLTHROUGH;
 		case 'd':
 			time *= 24;
-			fallthrough;
+			BFS_FALLTHROUGH;
 		case 'h':
 			time *= 60;
-			fallthrough;
+			BFS_FALLTHROUGH;
 		case 'm':
 			time *= 60;
-			fallthrough;
+			BFS_FALLTHROUGH;
 		case 's':
 			break;
 		default:
@@ -1189,6 +1246,10 @@ static struct expr *parse_exec(struct parser_state *state, int flags, int arg2)
 		}
 	}
 
+	if (execbuf->flags & BFS_EXEC_CONFIRM) {
+		state->ok_arg = expr->argv[0];
+	}
+
 	return expr;
 }
 
@@ -1232,6 +1293,58 @@ static struct expr *parse_f(struct parser_state *state, int arg1, int arg2) {
 	return &expr_true;
 }
 
+/**
+ * Parse -files0-from PATH.
+ */
+static struct expr *parse_files0_from(struct parser_state *state, int arg1, int arg2) {
+	const char *arg = state->argv[0];
+	const char *from = state->argv[1];
+	if (!from) {
+		parse_error(state, "${blu}%s${rs} requires a path.\n", arg);
+		return NULL;
+	}
+
+	FILE *file;
+	if (strcmp(from, "-") == 0) {
+		file = stdin;
+	} else {
+		file = xfopen(from, O_RDONLY | O_CLOEXEC);
+	}
+	if (!file) {
+		parse_error(state, "${blu}%s${rs} ${bld}%s${rs}: %m.\n", arg, from);
+		return NULL;
+	}
+
+	struct expr *expr = parse_unary_positional_option(state);
+
+	while (true) {
+		char *path = xgetdelim(file, '\0');
+		if (!path) {
+			if (errno) {
+				parse_error(state, "${blu}%s${rs} ${bld}%s${rs}: %m.\n", arg, from);
+				expr = NULL;
+			}
+			break;
+		}
+
+		int ret = parse_root(state, path);
+		free(path);
+		if (ret != 0) {
+			expr = NULL;
+			break;
+		}
+	}
+
+	if (file == stdin) {
+		state->stdin_consumed = true;
+	} else {
+		fclose(file);
+	}
+
+	state->implicit_root = false;
+	return expr;
+}
+
 /**
  * Parse -flags FLAGS.
  */
@@ -1873,7 +1986,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 		case MODE_CLAUSE:
 			who = 0;
 			mstate = MODE_WHO;
-			fallthrough;
+			BFS_FALLTHROUGH;
 
 		case MODE_WHO:
 			switch (*i) {
@@ -1900,7 +2013,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 			case MODE_EQUALS:
 				expr->file_mode &= ~who;
 				expr->dir_mode &= ~who;
-				fallthrough;
+				BFS_FALLTHROUGH;
 			case MODE_PLUS:
 				expr->file_mode |= file_change;
 				expr->dir_mode |= dir_change;
@@ -1910,7 +2023,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 				expr->dir_mode &= ~dir_change;
 				break;
 			}
-			fallthrough;
+			BFS_FALLTHROUGH;
 
 		case MODE_ACTION:
 			if (who == 0) {
@@ -1992,7 +2105,7 @@ static int parse_mode(const struct parser_state *state, const char *mode, struct
 				break;
 			case 'x':
 				file_change |= who & 0111;
-				fallthrough;
+				BFS_FALLTHROUGH;
 			case 'X':
 				dir_change |= who & 0111;
 				break;
@@ -2055,7 +2168,7 @@ static struct expr *parse_perm(struct parser_state *state, int field, int arg2)
 			++mode;
 			break;
 		}
-		fallthrough;
+		BFS_FALLTHROUGH;
 	default:
 		expr->mode_cmp = MODE_EXACT;
 		break;
@@ -2567,13 +2680,23 @@ static struct expr *parse_xdev(struct parser_state *state, int arg1, int arg2) {
  */
 static CFILE *launch_pager(pid_t *pid, CFILE *cout) {
 	char *pager = getenv("PAGER");
-	if (!pager || !pager[0]) {
-		pager = "more";
+
+	char *exe;
+	if (pager && pager[0]) {
+		exe = bfs_spawn_resolve(pager);
+	} else {
+		exe = bfs_spawn_resolve("less");
+		if (!exe) {
+			exe = bfs_spawn_resolve("more");
+		}
+	}
+	if (!exe) {
+		goto fail;
 	}
 
 	int pipefd[2];
 	if (pipe(pipefd) != 0) {
-		goto fail;
+		goto fail_exe;
 	}
 
 	FILE *file = fdopen(pipefd[1], "w");
@@ -2582,23 +2705,17 @@ static CFILE *launch_pager(pid_t *pid, CFILE *cout) {
 	}
 	pipefd[1] = -1;
 
-	CFILE *ret = cfdup(file, NULL);
+	CFILE *ret = cfwrap(file, NULL, true);
 	if (!ret) {
 		goto fail_file;
 	}
 	file = NULL;
-	ret->close = true;
-	ret->colors = cout->colors;
 
 	struct bfs_spawn ctx;
 	if (bfs_spawn_init(&ctx) != 0) {
 		goto fail_ret;
 	}
 
-	if (bfs_spawn_setflags(&ctx, BFS_SPAWN_USEPATH) != 0) {
-		goto fail_ctx;
-	}
-
 	if (bfs_spawn_addclose(&ctx, fileno(ret->file)) != 0) {
 		goto fail_ctx;
 	}
@@ -2610,49 +2727,27 @@ static CFILE *launch_pager(pid_t *pid, CFILE *cout) {
 	}
 
 	char *argv[] = {
-		pager,
+		exe,
+		NULL,
 		NULL,
 	};
 
-	extern char **environ;
-	char **envp = environ;
-
-	const char *less = getenv("LESS");
-	if (!less || !less[0]) {
-		size_t envc;
-		for (envc = 0; environ[envc]; ++envc) { }
-
-		envp = malloc((envc + 2)*sizeof(*envp));
-		if (!envp) {
-			goto fail_ctx;
-		}
-
-		size_t j = 0;
-		for (size_t i = 0; i < envc; ++i) {
-			if (strncmp(environ[i], "LESS=", 5) != 0) {
-				envp[j++] = environ[i];
-			}
-		}
-		envp[j++] = "LESS=FKRX";
-		envp[j] = NULL;
+	if (strcmp(xbasename(exe), "less") == 0) {
+		// We know less supports colors, other pagers may not
+		ret->colors = cout->colors;
+		argv[1] = "-FKRX";
 	}
 
-	*pid = bfs_spawn(pager, &ctx, argv, envp);
+	*pid = bfs_spawn(exe, &ctx, argv, NULL);
 	if (*pid < 0) {
-		goto fail_envp;
+		goto fail_ctx;
 	}
 
-	close(pipefd[0]);
-	if (envp != environ) {
-		free(envp);
-	}
+	xclose(pipefd[0]);
 	bfs_spawn_destroy(&ctx);
+	free(exe);
 	return ret;
 
-fail_envp:
-	if (envp != environ) {
-		free(envp);
-	}
 fail_ctx:
 	bfs_spawn_destroy(&ctx);
 fail_ret:
@@ -2663,11 +2758,13 @@ fail_file:
 	}
 fail_pipe:
 	if (pipefd[1] >= 0) {
-		close(pipefd[1]);
+		xclose(pipefd[1]);
 	}
 	if (pipefd[0] >= 0) {
-		close(pipefd[0]);
+		xclose(pipefd[0]);
 	}
+fail_exe:
+	free(exe);
 fail:
 	return cout;
 }
@@ -2751,6 +2848,8 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "      Measure times relative to the start of today\n");
 	cfprintf(cout, "  ${blu}-depth${rs}\n");
 	cfprintf(cout, "      Search in post-order (descendents first)\n");
+	cfprintf(cout, "  ${blu}-files0-from${rs} ${bld}FILE${rs}\n");
+	cfprintf(cout, "      Search the NUL ('\\0')-separated paths from ${bld}FILE${rs} (${bld}-${rs} for standard input).\n");
 	cfprintf(cout, "  ${blu}-follow${rs}\n");
 	cfprintf(cout, "      Follow all symbolic links (same as ${cyn}-L${rs})\n");
 	cfprintf(cout, "  ${blu}-ignore_readdir_race${rs}\n");
@@ -2897,7 +2996,7 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "  ${blu}-fls${rs} ${bld}FILE${rs}\n");
 	cfprintf(cout, "  ${blu}-fprint${rs} ${bld}FILE${rs}\n");
 	cfprintf(cout, "  ${blu}-fprint0${rs} ${bld}FILE${rs}\n");
-	cfprintf(cout, "  ${blu}-fprintf${rs} ${bld}FORMAT${rs} ${bld}FILE${rs}\n");
+	cfprintf(cout, "  ${blu}-fprintf${rs} ${bld}FILE${rs} ${bld}FORMAT${rs}\n");
 	cfprintf(cout, "      Like ${blu}-ls${rs}/${blu}-print${rs}/${blu}-print0${rs}/${blu}-printf${rs}, but write to ${bld}FILE${rs} instead of standard\n"
 	               "      output\n");
 	cfprintf(cout, "  ${blu}-ls${rs}\n");
@@ -3003,6 +3102,7 @@ static const struct table_entry parse_table[] = {
 	{"-exit", T_ACTION, parse_exit},
 	{"-f", T_FLAG, parse_f},
 	{"-false", T_TEST, parse_const, false},
+	{"-files0-from", T_OPTION, parse_files0_from},
 	{"-flags", T_TEST, parse_flags},
 	{"-fls", T_ACTION, parse_fls},
 	{"-follow", T_OPTION, parse_follow, BFTW_FOLLOW_ALL, true},
@@ -3423,6 +3523,11 @@ static struct expr *parse_whole_expr(struct parser_state *state) {
 		fprintf(stderr, "\n");
 	}
 
+	if (state->ok_arg && state->stdin_consumed) {
+		parse_error(state, "${blu}%s${rs} conflicts with ${blu}-files0-from${rs} ${bld}-${rs}.\n", state->ok_arg);
+		goto fail;
+	}
+
 	return expr;
 
 fail:
@@ -3613,15 +3718,20 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 		ctx->colors_error = errno;
 	}
 
-	ctx->cerr = cfdup(stderr, use_color ? ctx->colors : NULL);
+	ctx->cerr = cfwrap(stderr, use_color ? ctx->colors : NULL, false);
 	if (!ctx->cerr) {
-		perror("cfdup()");
+		perror("cfwrap()");
 		goto fail;
 	}
 
-	ctx->cout = cfdup(stdout, use_color ? ctx->colors : NULL);
+	ctx->cout = cfwrap(stdout, use_color ? ctx->colors : NULL, false);
 	if (!ctx->cout) {
-		bfs_perror(ctx, "cfdup()");
+		bfs_perror(ctx, "cfwrap()");
+		goto fail;
+	}
+
+	if (!bfs_ctx_dedup(ctx, ctx->cout, NULL) || !bfs_ctx_dedup(ctx, ctx->cerr, NULL)) {
+		bfs_perror(ctx, "bfs_ctx_dedup()");
 		goto fail;
 	}
 
@@ -3629,7 +3739,9 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 	bool stdout_tty = isatty(STDOUT_FILENO);
 	bool stderr_tty = isatty(STDERR_FILENO);
 
-	if (!getenv("POSIXLY_CORRECT")) {
+	if (getenv("POSIXLY_CORRECT")) {
+		ctx->posixly_correct = true;
+	} else {
 		ctx->warn = stdin_tty;
 	}
 
@@ -3640,8 +3752,10 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 		.regex_flags = 0,
 		.stdout_tty = stdout_tty,
 		.interactive = stdin_tty && stderr_tty,
+		.stdin_consumed = false,
 		.use_color = use_color,
 		.implicit_print = true,
+		.implicit_root = true,
 		.non_option_seen = false,
 		.just_info = false,
 		.excluding = false,
@@ -3650,6 +3764,7 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 		.prune_arg = NULL,
 		.mount_arg = NULL,
 		.xdev_arg = NULL,
+		.ok_arg = NULL,
 	};
 
 	if (strcmp(xbasename(state.command), "find") == 0) {
@@ -3676,7 +3791,10 @@ struct bfs_ctx *bfs_parse_cmdline(int argc, char *argv[]) {
 	}
 
 	if (darray_length(ctx->paths) == 0) {
-		if (parse_root(&state, ".") != 0) {
+		if (!state.implicit_root) {
+			parse_error(&state, "No root paths specified.\n");
+			goto fail;
+		} else if (parse_root(&state, ".") != 0) {
 			goto fail;
 		}
 	}
