@@ -1,4 +1,4 @@
-#!/bin/bash
+#!/usr/bin/env bash
 
 ############################################################################
 # bfs                                                                      #
@@ -21,7 +21,13 @@ set -o physical
 umask 022
 
 export LC_ALL=C
-export TZ=UTC
+export TZ=UTC0
+
+export ASAN_OPTIONS="abort_on_error=1"
+export LSAN_OPTIONS="abort_on_error=1"
+export MSAN_OPTIONS="abort_on_error=1"
+export TSAN_OPTIONS="abort_on_error=1"
+export UBSAN_OPTIONS="abort_on_error=1"
 
 if [ -t 1 ]; then
     BLD="$(printf '\033[01m')"
@@ -503,6 +509,14 @@ gnu_tests=(
 
     test_false
 
+    test_files0_from_file
+    test_files0_from_stdin
+    test_files0_from_none
+    test_files0_from_empty
+    test_files0_from_nowhere
+    test_files0_from_nothing
+    test_files0_from_ok
+
     test_fls
 
     test_follow
@@ -510,10 +524,16 @@ gnu_tests=(
     test_fprint
     test_fprint_duplicate
     test_fprint_error
+    test_fprint_noerror
+    test_fprint_noarg
+    test_fprint_nonexistent
+    test_fprint_truncate
 
     test_fprint0
 
     test_fprintf
+    test_fprintf_nofile
+    test_fprintf_noformat
 
     test_fstype
 
@@ -651,6 +671,8 @@ bfs_tests=(
     test_expr_flag_path
     test_expr_path_flag
 
+    test_unexpected_operator
+
     test_typo
 
     # Flags
@@ -715,11 +737,20 @@ bfs_tests=(
 
     test_execdir_plus
 
+    test_fprint_duplicate_stdout
+    test_fprint_error_stdout
+    test_fprint_error_stderr
+
     test_help
 
     test_hidden
     test_hidden_root
 
+    test_links_noarg
+    test_links_empty
+    test_links_negative
+    test_links_invalid
+
     test_newerma_nonexistent
     test_newermt_invalid
     test_newermq
@@ -740,6 +771,7 @@ bfs_tests=(
     test_printf_invalid_format
     test_printf_duplicate_flag
     test_printf_must_be_numeric
+    test_printf_color
 
     test_type_multi
 
@@ -760,6 +792,10 @@ bfs_tests=(
 
     # PATH_MAX handling
     test_deep_strict
+
+    # Error handling
+    test_stderr_fails_silently
+    test_stderr_fails_loudly
 )
 
 sudo_tests=(
@@ -849,6 +885,13 @@ function installp() {
     install "$@"
 }
 
+# Prefer GNU touch to work around https://apple.stackexchange.com/a/425730/397839
+if command -v gtouch &>/dev/null; then
+    TOUCH=gtouch
+else
+    TOUCH=touch
+fi
+
 # Like a mythical touch -p
 function touchp() {
     for arg; do
@@ -914,12 +957,12 @@ make_loops "$TMP/loops"
 # Creates a file+directory structure with varying timestamps
 function make_times() {
     mkdir -p "$1"
-    touch -t 199112140000 "$1/a"
-    touch -t 199112140001 "$1/b"
-    touch -t 199112140002 "$1/c"
+    $TOUCH -t 199112140000 "$1/a"
+    $TOUCH -t 199112140001 "$1/b"
+    $TOUCH -t 199112140002 "$1/c"
     ln -s a "$1/l"
-    touch -h -t 199112140003 "$1/l"
-    touch -t 199112140004 "$1"
+    $TOUCH -h -t 199112140003 "$1/l"
+    $TOUCH -t 199112140004 "$1"
 }
 make_times "$TMP/times"
 
@@ -959,7 +1002,7 @@ function make_deep() {
                 cd "$name" 2>/dev/null
             done
 
-            touch "$name"
+            $TOUCH "$name"
         )
     done
 }
@@ -1051,6 +1094,20 @@ function quiet() {
     fi
 }
 
+# Expect a command to fail, but not crash
+function fail() {
+    "$@"
+    local STATUS="$?"
+
+    if ((STATUS > 125)); then
+        return "$STATUS"
+    elif ((STATUS)); then
+        return 0
+    else
+        return 1
+    fi
+}
+
 # Return value when bfs fails
 EX_BFS=10
 # Return value when a difference is detected
@@ -1326,6 +1383,22 @@ function test_links_minus() {
     bfs_diff links -type f -links -2
 }
 
+function test_links_noarg() {
+    fail quiet invoke_bfs links -links
+}
+
+function test_links_empty() {
+    fail quiet invoke_bfs links -links ''
+}
+
+function test_links_negative() {
+    fail quiet invoke_bfs links -links +-1
+}
+
+function test_links_invalid() {
+    fail quiet invoke_bfs links -links ASDF
+}
+
 function test_P() {
     bfs_diff -P links/deeply/nested/dir
 }
@@ -1464,7 +1537,7 @@ function test_xtype_reorder() {
 
 function test_xtype_depth() {
     # Make sure -xtype is considered side-effecting for facts_when_impure
-    ! quiet invoke_bfs loops -xtype l -depth 100
+    fail quiet invoke_bfs loops -xtype l -depth 100
 }
 
 function test_iname() {
@@ -1538,19 +1611,19 @@ function test_newermt_epoch_minus_one() {
 }
 
 function test_newermt_invalid() {
-    ! quiet invoke_bfs times -newermt not_a_date_time
+    fail quiet invoke_bfs times -newermt not_a_date_time
 }
 
 function test_newerma_nonexistent() {
-    ! quiet invoke_bfs times -newerma basic/nonexistent
+    fail quiet invoke_bfs times -newerma basic/nonexistent
 }
 
 function test_newermq() {
-    ! quiet invoke_bfs times -newermq times/a
+    fail quiet invoke_bfs times -newermq times/a
 }
 
 function test_newerqm() {
-    ! quiet invoke_bfs times -newerqm times/a
+    fail quiet invoke_bfs times -newerqm times/a
 }
 
 function test_size() {
@@ -1575,7 +1648,7 @@ function test_exec() {
 
 function test_exec_nothing() {
     # Regression test: don't segfault on missing command
-    ! quiet invoke_bfs basic -exec \;
+    fail quiet invoke_bfs basic -exec \;
 }
 
 function test_exec_plus() {
@@ -1586,7 +1659,7 @@ function test_exec_plus_status() {
     # -exec ... {} + should always return true, but if the command fails, bfs
     # should exit with a non-zero status
     bfs_diff basic -exec false '{}' + -print
-    ! invoke_bfs basic -exec false '{}' +
+    (($? == EX_BFS))
 }
 
 function test_exec_plus_semicolon() {
@@ -1675,9 +1748,9 @@ function test_fprint() {
     sort -o scratch/test_fprint.out scratch/test_fprint.out
 
     if [ "$UPDATE" ]; then
-        cp scratch/test_fprint.out "$TESTS/test_fprint.out"
+        cp {scratch,"$TESTS"}/test_fprint.out
     else
-        diff -u scratch/test_fprint.out "$TESTS/test_fprint.out"
+        diff -u {"$TESTS",scratch}/test_fprint.out
     fi
 }
 
@@ -1690,9 +1763,43 @@ function test_fprint_duplicate() {
     sort -o scratch/test_fprint_duplicate.out scratch/test_fprint_duplicate.out
 
     if [ "$UPDATE" ]; then
-        cp scratch/test_fprint_duplicate.out "$TESTS/test_fprint_duplicate.out"
+        cp {scratch,"$TESTS"}/test_fprint_duplicate.out
     else
-        diff -u scratch/test_fprint_duplicate.out "$TESTS/test_fprint_duplicate.out"
+        diff -u {"$TESTS",scratch}/test_fprint_duplicate.out
+    fi
+}
+
+function test_fprint_duplicate_stdout() {
+    touchp scratch/test_fprint_duplicate_stdout.out
+
+    invoke_bfs basic -fprint scratch/test_fprint_duplicate_stdout.out -print >scratch/test_fprint_duplicate_stdout.out
+    sort -o scratch/test_fprint_duplicate_stdout.out{,}
+
+    if [ "$UPDATE" ]; then
+        cp {scratch,"$TESTS"}/test_fprint_duplicate_stdout.out
+    else
+        diff -u {"$TESTS",scratch}/test_fprint_duplicate_stdout.out
+    fi
+}
+
+function test_fprint_noarg() {
+    fail quiet invoke_bfs basic -fprint
+}
+
+function test_fprint_nonexistent() {
+    fail quiet invoke_bfs basic -fprint scratch/nonexistent/path
+}
+
+function test_fprint_truncate() {
+    printf "basic\nbasic\n" >scratch/test_fprint_truncate.out
+
+    invoke_bfs basic -maxdepth 0 -fprint scratch/test_fprint_truncate.out
+    sort -o scratch/test_fprint_truncate.out scratch/test_fprint_truncate.out
+
+    if [ "$UPDATE" ]; then
+        cp {scratch,"$TESTS"}/test_fprint_truncate.out
+    else
+        diff -u {"$TESTS",scratch}/test_fprint_truncate.out
     fi
 }
 
@@ -1708,7 +1815,7 @@ function test_flag_double_dash() {
 
 function test_ignore_readdir_race() {
     rm -rf scratch/*
-    touch scratch/{foo,bar}
+    $TOUCH scratch/{foo,bar}
 
     # -links 1 forces a stat() call, which will fail for the second file
     invoke_bfs scratch -mindepth 1 -ignore_readdir_race -links 1 -exec "$TESTS/remove-sibling.sh" '{}' ';'
@@ -1716,7 +1823,7 @@ function test_ignore_readdir_race() {
 
 function test_ignore_readdir_race_root() {
     # Make sure -ignore_readdir_race doesn't suppress ENOENT at the root
-    ! quiet invoke_bfs basic/nonexistent -ignore_readdir_race
+    fail quiet invoke_bfs basic/nonexistent -ignore_readdir_race
 }
 
 function test_ignore_readdir_race_notdir() {
@@ -1724,7 +1831,7 @@ function test_ignore_readdir_race_notdir() {
     rm -rf scratch/*
     touchp scratch/foo/bar
 
-    invoke_bfs scratch -mindepth 1 -ignore_readdir_race -execdir rm -r '{}' \; -execdir touch '{}' \;
+    invoke_bfs scratch -mindepth 1 -ignore_readdir_race -execdir rm -r '{}' \; -execdir $TOUCH '{}' \;
 }
 
 function test_perm_000() {
@@ -1788,15 +1895,15 @@ function test_perm_symbolic_slash() {
 }
 
 function test_perm_symbolic_trailing_comma() {
-    ! quiet invoke_bfs perms -perm a+r,
+    fail quiet invoke_bfs perms -perm a+r,
 }
 
 function test_perm_symbolic_double_comma() {
-    ! quiet invoke_bfs perms -perm a+r,,u+w
+    fail quiet invoke_bfs perms -perm a+r,,u+w
 }
 
 function test_perm_symbolic_missing_action() {
-    ! quiet invoke_bfs perms -perm a
+    fail quiet invoke_bfs perms -perm a
 }
 
 function test_perm_leading_plus_symbolic() {
@@ -1837,7 +1944,7 @@ function test_not_prune() {
 
 function test_ok_nothing() {
     # Regression test: don't segfault on missing command
-    ! quiet invoke_bfs basic -ok \;
+    fail quiet invoke_bfs basic -ok \;
 }
 
 function test_ok_stdin() {
@@ -1874,7 +1981,7 @@ function test_delete_many() {
 
     rm -rf scratch/*
     mkdir scratch/foo
-    touch scratch/foo/{1..256}
+    $TOUCH scratch/foo/{1..256}
 
     invoke_bfs scratch/foo -delete
     bfs_diff scratch
@@ -1915,7 +2022,7 @@ function test_regex_parens() {
 }
 
 function test_regex_error() {
-    ! quiet invoke_bfs basic -regex '['
+    fail quiet invoke_bfs basic -regex '['
 }
 
 function test_E() {
@@ -2189,27 +2296,31 @@ function test_printf_l_nonlink() {
 }
 
 function test_printf_incomplete_escape() {
-    ! quiet invoke_bfs basic -printf '\'
+    fail quiet invoke_bfs basic -printf '\'
 }
 
 function test_printf_invalid_escape() {
-    ! quiet invoke_bfs basic -printf '\!'
+    fail quiet invoke_bfs basic -printf '\!'
 }
 
 function test_printf_incomplete_format() {
-    ! quiet invoke_bfs basic -printf '%'
+    fail quiet invoke_bfs basic -printf '%'
 }
 
 function test_printf_invalid_format() {
-    ! quiet invoke_bfs basic -printf '%!'
+    fail quiet invoke_bfs basic -printf '%!'
 }
 
 function test_printf_duplicate_flag() {
-    ! quiet invoke_bfs basic -printf '%--p'
+    fail quiet invoke_bfs basic -printf '%--p'
 }
 
 function test_printf_must_be_numeric() {
-    ! quiet invoke_bfs basic -printf '%+p'
+    fail quiet invoke_bfs basic -printf '%+p'
+}
+
+function test_printf_color() {
+    LS_COLORS= bfs_diff -color -path './rainbow*' -printf '%H %h %f %p %P %l\n'
 }
 
 function test_fprintf() {
@@ -2219,10 +2330,18 @@ function test_fprintf() {
     if [ "$UPDATE" ]; then
         cp scratch/test_fprintf.out "$TESTS/test_fprintf.out"
     else
-        diff -u scratch/test_fprintf.out "$TESTS/test_fprintf.out"
+        diff -u "$TESTS/test_fprintf.out" scratch/test_fprintf.out
     fi
 }
 
+function test_fprintf_nofile() {
+    fail quiet invoke_bfs basic -fprintf
+}
+
+function test_fprintf_noformat() {
+    fail quiet invoke_bfs basic -fprintf /dev/null
+}
+
 function test_fstype() {
     fstype="$(invoke_bfs basic -maxdepth 0 -printf '%F\n')"
     bfs_diff basic -fstype "$fstype"
@@ -2289,15 +2408,15 @@ function test_precedence() {
 }
 
 function test_incomplete() {
-    ! quiet invoke_bfs basic \(
+    fail quiet invoke_bfs basic \(
 }
 
 function test_missing_paren() {
-    ! quiet invoke_bfs basic \( -print
+    fail quiet invoke_bfs basic \( -print
 }
 
 function test_extra_paren() {
-    ! quiet invoke_bfs basic -print \)
+    fail quiet invoke_bfs basic -print \)
 }
 
 function test_color() {
@@ -2588,13 +2707,33 @@ function test_data_flow_or_swap() {
 
 function test_print_error() {
     if [ -e /dev/full ]; then
-        ! quiet invoke_bfs basic -maxdepth 0 >/dev/full
+        fail quiet invoke_bfs basic -maxdepth 0 >/dev/full
     fi
 }
 
 function test_fprint_error() {
     if [ -e /dev/full ]; then
-        ! quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full
+        fail quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full
+    fi
+}
+
+function test_fprint_noerror() {
+    # Regression test: /dev/full should not fail until actually written to
+
+    if [ -e /dev/full ]; then
+        invoke_bfs basic -false -fprint /dev/full
+    fi
+}
+
+function test_fprint_error_stdout() {
+    if [ -e /dev/full ]; then
+        fail quiet invoke_bfs basic -maxdepth 0 -fprint /dev/full >/dev/full
+    fi
+}
+
+function test_fprint_error_stderr() {
+    if [ -e /dev/full ]; then
+        fail invoke_bfs basic -maxdepth 0 -fprint /dev/full 2>/dev/full
     fi
 }
 
@@ -2631,11 +2770,11 @@ function test_okdir_closed_stdin() {
 }
 
 function test_closed_stdout() {
-    ! quiet invoke_bfs basic >&-
+    fail quiet invoke_bfs basic >&-
 }
 
 function test_closed_stderr() {
-    ! invoke_bfs basic >&- 2>&-
+    fail invoke_bfs basic >&- 2>&-
 }
 
 function test_unique() {
@@ -2662,7 +2801,7 @@ function test_mount() {
     rm -rf scratch/*
     mkdir scratch/{foo,mnt}
     sudo mount -t tmpfs tmpfs scratch/mnt
-    touch scratch/foo/bar scratch/mnt/baz
+    $TOUCH scratch/foo/bar scratch/mnt/baz
 
     bfs_diff scratch -mount
     local ret=$?
@@ -2676,7 +2815,7 @@ function test_L_mount() {
     mkdir scratch/{foo,mnt}
     sudo mount -t tmpfs tmpfs scratch/mnt
     ln -s ../mnt scratch/foo/bar
-    touch scratch/mnt/baz
+    $TOUCH scratch/mnt/baz
     ln -s ../mnt/baz scratch/foo/qux
 
     bfs_diff -L scratch -mount
@@ -2690,7 +2829,7 @@ function test_xdev() {
     rm -rf scratch/*
     mkdir scratch/{foo,mnt}
     sudo mount -t tmpfs tmpfs scratch/mnt
-    touch scratch/foo/bar scratch/mnt/baz
+    $TOUCH scratch/foo/bar scratch/mnt/baz
 
     bfs_diff scratch -xdev
     local ret=$?
@@ -2704,7 +2843,7 @@ function test_L_xdev() {
     mkdir scratch/{foo,mnt}
     sudo mount -t tmpfs tmpfs scratch/mnt
     ln -s ../mnt scratch/foo/bar
-    touch scratch/mnt/baz
+    $TOUCH scratch/mnt/baz
     ln -s ../mnt/baz scratch/foo/qux
 
     bfs_diff -L scratch -xdev
@@ -2728,7 +2867,7 @@ function test_inum_mount() {
 
 function test_inum_bind_mount() {
     rm -rf scratch/*
-    touch scratch/{foo,bar}
+    $TOUCH scratch/{foo,bar}
     sudo mount --bind scratch/{foo,bar}
 
     bfs_diff scratch -inum "$(inum scratch/bar)"
@@ -2740,7 +2879,7 @@ function test_inum_bind_mount() {
 
 function test_type_bind_mount() {
     rm -rf scratch/*
-    touch scratch/{file,null}
+    $TOUCH scratch/{file,null}
     sudo mount --bind /dev/null scratch/null
 
     bfs_diff scratch -type c
@@ -2752,7 +2891,7 @@ function test_type_bind_mount() {
 
 function test_xtype_bind_mount() {
     rm -rf scratch/*
-    touch scratch/{file,null}
+    $TOUCH scratch/{file,null}
     sudo mount --bind /dev/null scratch/null
     ln -s /dev/null scratch/link
 
@@ -2786,7 +2925,7 @@ function test_acl() {
 
     quiet invoke_bfs scratch -quit -acl || return 0
 
-    touch scratch/{normal,acl}
+    $TOUCH scratch/{normal,acl}
     set_acl scratch/acl || return 0
     ln -s acl scratch/link
 
@@ -2798,7 +2937,7 @@ function test_L_acl() {
 
     quiet invoke_bfs scratch -quit -acl || return 0
 
-    touch scratch/{normal,acl}
+    $TOUCH scratch/{normal,acl}
     set_acl scratch/acl || return 0
     ln -s acl scratch/link
 
@@ -2808,11 +2947,11 @@ function test_L_acl() {
 function test_capable() {
     rm -rf scratch/*
 
-    if ! quiet invoke_bfs scratch -quit -capable; then
+    if fail quiet invoke_bfs scratch -quit -capable; then
         return 0
     fi
 
-    touch scratch/{normal,capable}
+    $TOUCH scratch/{normal,capable}
     sudo setcap all+ep scratch/capable
     ln -s capable scratch/link
 
@@ -2822,11 +2961,11 @@ function test_capable() {
 function test_L_capable() {
     rm -rf scratch/*
 
-    if ! quiet invoke_bfs scratch -quit -capable; then
+    if fail quiet invoke_bfs scratch -quit -capable; then
         return 0
     fi
 
-    touch scratch/{normal,capable}
+    $TOUCH scratch/{normal,capable}
     sudo setcap all+ep scratch/capable
     ln -s capable scratch/link
 
@@ -2836,7 +2975,7 @@ function test_L_capable() {
 function make_xattrs() {
     rm -rf scratch/*
 
-    touch scratch/{normal,xattr,xattr_2}
+    $TOUCH scratch/{normal,xattr,xattr_2}
     ln -s xattr scratch/link
     ln -s normal scratch/xattr_link
 
@@ -2981,11 +3120,11 @@ function test_exclude_mindepth() {
 }
 
 function test_exclude_print() {
-    ! quiet invoke_bfs basic -exclude -print
+    fail quiet invoke_bfs basic -exclude -print
 }
 
 function test_exclude_exclude() {
-    ! quiet invoke_bfs basic -exclude -exclude -name foo
+    fail quiet invoke_bfs basic -exclude -exclude -name foo
 }
 
 function test_flags() {
@@ -2993,12 +3132,58 @@ function test_flags() {
 
     rm -rf scratch/*
 
-    touch scratch/{foo,bar}
+    $TOUCH scratch/{foo,bar}
     chflags offline scratch/bar || return 0
 
     bfs_diff scratch -flags -offline,nohidden
 }
 
+function test_files0_from_file() {
+    cd weirdnames
+    invoke_bfs -mindepth 1 -fprintf ../scratch/files0.in "%P\0"
+    bfs_diff -files0-from ../scratch/files0.in
+}
+
+function test_files0_from_stdin() {
+    cd weirdnames
+    invoke_bfs -mindepth 1 -printf "%P\0" | bfs_diff -files0-from -
+}
+
+function test_files0_from_none() {
+    printf "" | fail quiet invoke_bfs -files0-from -
+}
+
+function test_files0_from_empty() {
+    printf "\0" | fail quiet invoke_bfs -files0-from -
+}
+
+function test_files0_from_nowhere() {
+    fail quiet invoke_bfs -files0-from
+}
+
+function test_files0_from_nothing() {
+    fail quiet invoke_bfs -files0-from basic/nonexistent
+}
+
+function test_files0_from_ok() {
+    printf "basic\0" | fail quiet invoke_bfs -files0-from - -ok echo {} \;
+}
+
+function test_stderr_fails_silently() {
+    if [ -e /dev/full ]; then
+        bfs_diff -D all basic 2>/dev/full
+    fi
+}
+
+function test_stderr_fails_loudly() {
+    if [ -e /dev/full ]; then
+        fail invoke_bfs -D all basic -false -fprint /dev/full 2>/dev/full
+    fi
+}
+
+function test_unexpected_operator() {
+    fail quiet invoke_bfs \! -o -print
+}
 
 BOL=
 EOL='\n'
