@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2016-2021 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2016-2022 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -16,6 +16,7 @@
 
 #include "util.h"
 #include "dstring.h"
+#include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <langinfo.h>
@@ -88,7 +89,7 @@ int dup_cloexec(int fd) {
 	}
 
 	if (fcntl(ret, F_SETFD, FD_CLOEXEC) == -1) {
-		close(ret);
+		close_quietly(ret);
 		return -1;
 	}
 
@@ -105,10 +106,8 @@ int pipe_cloexec(int pipefd[2]) {
 	}
 
 	if (fcntl(pipefd[0], F_SETFD, FD_CLOEXEC) == -1 || fcntl(pipefd[1], F_SETFD, FD_CLOEXEC) == -1) {
-		int error = errno;
-		close(pipefd[1]);
-		close(pipefd[0]);
-		errno = error;
+		close_quietly(pipefd[1]);
+		close_quietly(pipefd[0]);
 		return -1;
 	}
 
@@ -266,31 +265,6 @@ bool is_nonexistence_error(int error) {
 	return error == ENOENT || errno == ENOTDIR;
 }
 
-/** Read a line from standard input. */
-static char *xgetline(void) {
-	char *line = dstralloc(0);
-	if (!line) {
-		return NULL;
-	}
-
-	while (true) {
-		int c = fgetc(stdin);
-		if (c == '\n' || c == EOF) {
-			break;
-		}
-
-		if (dstrapp(&line, c) != 0) {
-			goto error;
-		}
-	}
-
-	return line;
-
-error:
-	dstrfree(line);
-	return NULL;
-}
-
 /** Compile and execute a regular expression for xrpmatch(). */
 static int xrpregex(nl_item item, const char *response) {
 	const char *pattern = nl_langinfo(item);
@@ -336,12 +310,12 @@ static int xrpmatch(const char *response) {
 	}
 }
 
-int ynprompt() {
+int ynprompt(void) {
 	fflush(stderr);
 
-	char *line = xgetline();
+	char *line = xgetdelim(stdin, '\n');
 	int ret = line ? xrpmatch(line) : -1;
-	dstrfree(line);
+	free(line);
 	return ret;
 }
 
@@ -414,3 +388,97 @@ size_t xwrite(int fd, const void *buf, size_t nbytes) {
 
 	return count;
 }
+
+char *xconfstr(int name) {
+	size_t len = confstr(name, NULL, 0);
+	if (len == 0) {
+		return NULL;
+	}
+
+	char *str = malloc(len);
+	if (!str) {
+		return NULL;
+	}
+
+	if (confstr(name, str, len) != len) {
+		free(str);
+		return NULL;
+	}
+
+	return str;
+}
+
+char *xgetdelim(FILE *file, char delim) {
+	char *chunk = NULL;
+	size_t n = 0;
+	ssize_t len = getdelim(&chunk, &n, delim, file);
+	if (len >= 0) {
+		if (chunk[len] == delim) {
+			chunk[len] = '\0';
+		}
+		return chunk;
+	} else {
+		free(chunk);
+		if (!ferror(file)) {
+			errno = 0;
+		}
+		return NULL;
+	}
+}
+
+FILE *xfopen(const char *path, int flags) {
+	char mode[4];
+
+	switch (flags & O_ACCMODE) {
+	case O_RDONLY:
+		strcpy(mode, "rb");
+		break;
+	case O_WRONLY:
+		strcpy(mode, "wb");
+		break;
+	case O_RDWR:
+		strcpy(mode, "r+b");
+		break;
+	default:
+		assert(!"Invalid access mode");
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (flags & O_APPEND) {
+		mode[0] = 'a';
+	}
+
+	int fd;
+	if (flags & O_CREAT) {
+		fd = open(path, flags, 0666);
+	} else {
+		fd = open(path, flags);
+	}
+
+	if (fd < 0) {
+		return NULL;
+	}
+
+	FILE *ret = fdopen(fd, mode);
+	if (!ret) {
+		close_quietly(fd);
+		return NULL;
+	}
+
+	return ret;
+}
+
+int xclose(int fd) {
+	int ret = close(fd);
+	if (ret != 0) {
+		assert(errno != EBADF);
+	}
+	return ret;
+}
+
+void close_quietly(int fd) {
+	int error = errno;
+	xclose(fd);
+	errno = error;
+}
