@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2016-2021 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2016-2022 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -26,6 +26,7 @@
 #include <regex.h>
 #include <stdbool.h>
 #include <stddef.h>
+#include <stdio.h>
 #include <sys/types.h>
 
 // Some portability concerns
@@ -48,6 +49,12 @@
 #	define BFS_HAS_C_ATTRIBUTE(attr) false
 #endif
 
+#if __GNUC__ && defined(__has_attribute)
+#	define BFS_HAS_GNU_ATTRIBUTE(attr) __has_attribute(attr)
+#else
+#	define BFS_HAS_GNU_ATTRIBUTE(attr) false
+#endif
+
 #ifndef BFS_HAS_MNTENT
 #	define BFS_HAS_MNTENT BFS_HAS_INCLUDE(<mntent.h>, __GLIBC__)
 #endif
@@ -93,17 +100,17 @@
 #endif
 
 #if BFS_HAS_C_ATTRIBUTE(fallthrough)
-#	define fallthrough [[fallthrough]]
-#elif __GNUC__
-#	define fallthrough __attribute__((fallthrough))
+#	define BFS_FALLTHROUGH [[fallthrough]]
+#elif BFS_HAS_GNU_ATTRIBUTE(fallthrough)
+#	define BFS_FALLTHROUGH __attribute__((fallthrough))
 #else
-#	define fallthrough
+#	define BFS_FALLTHROUGH ((void)0)
 #endif
 
 /**
  * Adds compiler warnings for bad printf()-style function calls, if supported.
  */
-#if __GNUC__
+#if BFS_HAS_GNU_ATTRIBUTE(format)
 #	define BFS_FORMATTER(fmt, args) __attribute__((format(printf, fmt, args)))
 #else
 #	define BFS_FORMATTER(fmt, args)
@@ -258,4 +265,55 @@ size_t xread(int fd, void *buf, size_t nbytes);
  */
 size_t xwrite(int fd, const void *buf, size_t nbytes);
 
+/**
+ * Wrapper for confstr() that allocates with malloc().
+ *
+ * @param name
+ *         The ID of the confstr to look up.
+ * @return
+ *         The value of the confstr, or NULL on failure.
+ */
+char *xconfstr(int name);
+
+/**
+ * Convenience wrapper for getdelim().
+ *
+ * @param file
+ *         The file to read.
+ * @param delim
+ *         The delimiter character to split on.
+ * @return
+ *         The read chunk (without the delimiter), allocated with malloc().
+ *         NULL is returned on error (errno != 0) or end of file (errno == 0).
+ */
+char *xgetdelim(FILE *file, char delim);
+
+/**
+ * fopen() variant that takes open() style flags.
+ *
+ * @param path
+ *         The path to open.
+ * @param flags
+ *         Flags to pass to open().
+ */
+FILE *xfopen(const char *path, int flags);
+
+/**
+ * close() wrapper that asserts the file descriptor is valid.
+ *
+ * @param fd
+ *         The file descriptor to close.
+ * @return
+ *         0 on success, or -1 on error.
+ */
+int xclose(int fd);
+
+/**
+ * close() variant that preserves errno.
+ *
+ * @param fd
+ *         The file descriptor to close.
+ */
+void close_quietly(int fd);
+
 #endif // BFS_UTIL_H
